@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* job.c: job control */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -321,9 +321,14 @@ void apply_curstop(void)
 size_t job_count(void)
 {
     size_t count = 0;
-    for (size_t i = 0; i < joblist.length; i++)
-	if (joblist.contents[i] != NULL)
-	    count++;
+    for (size_t i = 0; i < joblist.length; i++) {
+	const job_T *job = joblist.contents[i];
+	if (job == NULL)
+	    continue;
+	if (job->j_status == JS_DONE && !job->j_statuschanged)
+	    continue; // Ignore finished jobs that have already been reported.
+	count++;
+    }
     return count;
 }
 
@@ -369,8 +374,8 @@ start:
 		    waitpidoption &= ~WCONTINUED;
 		    goto start;
 		}
-		/* falls thru! */
 #endif
+		; /* falls thru! */
 	    default:
 		xerror(errno, "waitpid");
 		return;
@@ -726,6 +731,18 @@ char *get_job_status_string(const job_T *job, bool *needfree)
     assert(false);
 }
 
+/* Returns true iff there is any job whose status has been changed but not yet
+ * reported. If this function returns false, `print_job_status_all' is nop. */
+bool any_job_status_has_changed(void)
+{
+    for (size_t i = 1; i < joblist.length; i++) {
+	const job_T *job = get_job(i);
+	if (job != NULL && !job->j_nonotify && job->j_statuschanged)
+	    return true;
+    }
+    return false;
+}
+
 /* Prints the status of the specified job.
  * If `remove_done' is true, finished jobs are removed from the job list after
  * the status is printed.
@@ -744,7 +761,7 @@ int print_job_status(size_t jobnumber,
     if (job == NULL || job->j_nonotify)
 	return result;
     if (changedonly && !job->j_statuschanged)
-	goto done;
+	return result;
 
     char current;
     if      (jobnumber == current_jobnumber)  current = '+';
@@ -800,7 +817,6 @@ int print_job_status(size_t jobnumber,
 	}
     }
     job->j_statuschanged = false;
-done:
     if (remove_done && job->j_status == JS_DONE)
 	remove_job(jobnumber);
 
@@ -1347,17 +1363,15 @@ int wait_for_job_by_jobspec(const wchar_t *jobspec)
 bool wait_builtin_has_job(bool jobcontrol)
 {
     /* print/remove already-finished jobs */
-    if (jobcontrol && is_interactive_now && !posixly_correct) {
-	for (size_t i = 1; i < joblist.length; i++)
-	    print_job_status(i, true, false, true, stdout);
-    } else {
-	for (size_t i = 1; i < joblist.length; i++) {
-	    job_T *job = joblist.contents[i];
-	    if (job != NULL && (job->j_legacy || job->j_status == JS_DONE))
-		remove_job(i);
-	}
+    for (size_t i = 1; i < joblist.length; i++) {
+	job_T *job = joblist.contents[i];
+	if (jobcontrol && is_interactive_now && !posixly_correct)
+	    print_job_status(i, true, false, false, stdout);
+	if (job != NULL && (job->j_legacy || job->j_status == JS_DONE))
+	    remove_job(i);
     }
 
+    /* see if we have jobs to wait for. */
     for (size_t i = 1; i < joblist.length; i++) {
 	job_T *job = joblist.contents[i];
 	if (job != NULL && (!jobcontrol || job->j_status == JS_RUNNING))
