@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* sig.c: signal handling */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2018 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -654,22 +654,18 @@ enum wait_for_input_T wait_for_input(int fd, bool trap, int timeout)
 /* Handles SIGCHLD if caught. */
 void handle_sigchld(void)
 {
-    static bool print_status = false;
-
     if (sigchld_received) {
         sigchld_received = false;
         do_wait();
-	print_status = true;
     }
 
-    if (print_status) {
+    if (any_job_status_has_changed()) {
 	/* print job status if the notify option is set */
 #if YASH_ENABLE_LINEEDIT
 	if (le_state & LE_STATE_ACTIVE) {
 	    if (!(le_state & LE_STATE_COMPLETING)) {
 		if (shopt_notify || shopt_notifyle) {
 		    le_suspend_readline();
-		    print_status = false;
 		    print_job_status_all();
 		    le_resume_readline();
 		}
@@ -682,7 +678,6 @@ void handle_sigchld(void)
 	    sigaddset(&ss, SIGTTOU);
 	    sigemptyset(&savess);
 	    sigprocmask(SIG_BLOCK, &ss, &savess);
-	    print_status = false;
 	    print_job_status_all();
 	    sigprocmask(SIG_SETMASK, &savess, NULL);
 	}
@@ -1254,7 +1249,7 @@ bool print_trap(const wchar_t *signame, const wchar_t *command)
     if (command == NULL)
 	return true;
 
-    wchar_t *q = quote_sq(command);
+    wchar_t *q = quote_as_word(command);
     bool ok = xprintf("trap -- %ls %ls\n", q, signame);
     free(q);
     return ok;
