@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* variable.c: deals with shell variables and parameters */
-/* (C) 2007-2015 magicant */
+/* (C) 2007-2018 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -230,6 +230,11 @@ void varkvfree(kvpair_T kv)
  * calls `varkvfree'. */
 void varkvfree_reexport(kvpair_T kv)
 {
+    if (kv.key == NULL) {
+	assert(kv.value == NULL);
+	return;
+    }
+
     variable_set(kv.key, NULL);
     if (((variable_T *) kv.value)->v_type & VF_EXPORT)
 	update_environment(kv.key);
@@ -336,8 +341,9 @@ void init_variables(void)
     }
 
     /* set $YASH_LOADPATH */
-    set_variable(L VAR_YASH_LOADPATH, xwcsdup(L DEFAULT_LOADPATH),
-	    SCOPE_GLOBAL, false);
+    if (getvar(L VAR_YASH_LOADPATH) == NULL)
+	set_variable(L VAR_YASH_LOADPATH, xwcsdup(L DEFAULT_LOADPATH),
+		SCOPE_GLOBAL, false);
 
     /* set $YASH_VERSION */
     set_variable(L VAR_YASH_VERSION, xwcsdup(L PACKAGE_VERSION),
@@ -734,10 +740,10 @@ bool do_assignments(const assign_T *assign, bool temp, bool export)
 
 	switch (assign->a_type) {
 	    case A_SCALAR:
-		value = expand_single(assign->a_scalar, TT_MULTI);
+		value = expand_single_and_unescape(
+			assign->a_scalar, TT_MULTI, true, false);
 		if (value == NULL)
 		    return false;
-		value = unescapefree(value);
 		if (shopt_xtrace)
 		    xtrace_variable(assign->a_name, value);
 		if (!set_variable(assign->a_name, value, scope, export))
@@ -762,7 +768,10 @@ bool do_assignments(const assign_T *assign, bool temp, bool export)
 void xtrace_variable(const wchar_t *name, const wchar_t *value)
 {
     xwcsbuf_T *buf = get_xtrace_buffer();
-    wb_wprintf(buf, L" %ls=%ls", name, value);
+    wb_wccat(buf, L' ');
+    wb_cat(buf, name);
+    wb_wccat(buf, L'=');
+    wb_quote_as_word(buf, value);
 }
 
 /* Pushes a trace of the specified array assignment to the xtrace buffer. */
@@ -773,7 +782,7 @@ void xtrace_array(const wchar_t *name, void *const *values)
     wb_wprintf(buf, L" %ls=(", name);
     if (*values != NULL) {
 	for (;;) {
-	    wb_cat(buf, *values);
+	    wb_quote_as_word(buf, *values);
 	    values++;
 	    if (*values == NULL)
 		break;
@@ -1639,29 +1648,43 @@ int typeset_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
-    if (wcscmp(ARGV(0), L"export") == 0) {
-	global = true;
-	if (!unexport)
-	    export = true;
-    } else if (wcscmp(ARGV(0), L"readonly") == 0) {
-	global = readonly = true;
-    } else {
-	assert(wcscmp(ARGV(0), L"typeset") == 0
-	    || wcscmp(ARGV(0), L"set") == 0);
+    switch (ARGV(0)[0]) {
+	case L'e':
+	    assert(wcscmp(ARGV(0), L"export") == 0);
+	    global = true;
+	    if (!unexport)
+		export = true;
+	    break;
+	case L'r':
+	    assert(wcscmp(ARGV(0), L"readonly") == 0);
+	    global = readonly = true;
+	    break;
+	case L's':
+	    assert(wcscmp(ARGV(0), L"set") == 0);
+	    global = true;
+	    break;
+	case L't':
+	    assert(wcscmp(ARGV(0), L"typeset") == 0);
+	    break;
+	default:
+	    assert(false);
     }
 
+    if (function && global && ARGV(0)[0] == L't' /*typeset*/)
+	return special_builtin_error(
+		mutually_exclusive_option_error(L'f', L'g'));
     if (function && export)
-	return special_builtin_syntax_error(
+	return special_builtin_error(
 		mutually_exclusive_option_error(L'f', L'x'));
     if (function && unexport)
-	return special_builtin_syntax_error(
+	return special_builtin_error(
 		mutually_exclusive_option_error(L'f', L'X'));
     if (export && unexport)
-	return special_builtin_syntax_error(
+	return special_builtin_error(
 		mutually_exclusive_option_error(L'x', L'X'));
 
     if (xoptind == argc) {
@@ -1728,10 +1751,13 @@ int typeset_builtin(int argc, void **argv)
 		/* treat function */
 		function_T *f = ht_get(&functions, arg).value;
 		if (f != NULL) {
-		    if (readonly)
-			f->f_type |= VF_READONLY | VF_NODELETE;
-		    if (print)
-			print_function(arg, f, ARGV(0), readonly);
+		    if (print) {
+			if (!readonly || (f->f_type & VF_READONLY))
+			    print_function(arg, f, ARGV(0), readonly);
+		    } else {
+			if (readonly)
+			    f->f_type |= VF_READONLY | VF_NODELETE;
+		    }
 		} else {
 		    xerror(0, Ngt("no such function `%ls'"), arg);
 		}
@@ -1739,7 +1765,8 @@ int typeset_builtin(int argc, void **argv)
 	} while (++xoptind < argc);
     }
 
-    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
+    return (yash_error_message_count == 0) ?
+	    Exit_SUCCESS : special_builtin_error(Exit_FAILURE);
 }
 
 /* Prints the specified variable to the standard output.
@@ -1762,7 +1789,7 @@ void print_variable(
 	return;
 
     if (!is_name(name))
-	name = qname = quote_sq(name);
+	name = qname = quote_as_word(name);
 
     switch (var->v_type & VF_MASK) {
 	case VF_SCALAR:
@@ -1777,8 +1804,9 @@ void print_variable(
 }
 
 /* Prints the specified scalar variable to the standard output.
- * If `is_name(name)' is not true, the quoted name must be given as `qname';
- * otherwise, `qname' must be NULL.
+ * `namequote' must equal `is_name(name)'. If `argv0' is "set" and `namequote'
+ * is true, the variable is not printed since it cannot be assigned in the
+ * normal assignment syntax.
  * An error message is printed to the standard error on error. */
 void print_scalar(const wchar_t *name, bool namequote,
 	const variable_T *var, const wchar_t *argv0)
@@ -1788,7 +1816,7 @@ void print_scalar(const wchar_t *name, bool namequote,
     xstrbuf_T opts;
 
     if (var->v_value != NULL)
-	quotedvalue = quote_sq(var->v_value);
+	quotedvalue = quote_as_word(var->v_value);
     else
 	quotedvalue = NULL;
     switch (argv0[0]) {
@@ -1834,7 +1862,7 @@ void print_array(
 	return;
     if (var->v_valc > 0) {
 	for (size_t i = 0; ; ) {
-	    wchar_t *qvalue = quote_sq(var->v_vals[i]);
+	    wchar_t *qvalue = quote_as_word(var->v_vals[i]);
 	    bool ok = xprintf("%ls", qvalue);
 	    free(qvalue);
 	    if (!ok)
@@ -1891,7 +1919,7 @@ void print_function(
 
     wchar_t *qname = NULL;
     if (!is_name(name))
-	name = qname = quote_sq(name);
+	name = qname = quote_as_word(name);
 
     wchar_t *value = command_to_wcs(func->f_body, true);
     const char *format = (qname == NULL) ? "%ls()\n%ls" : "function %ls()\n%ls";
@@ -2243,7 +2271,7 @@ int unset_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
@@ -2255,15 +2283,14 @@ int unset_builtin(int argc, void **argv)
 	if (function) {
 	    unset_function(name);
 	} else {
-	    if (wcschr(name, L'=')) {
-		xerror(0, Ngt("`%ls' is not a valid variable name"), name);
+	    if (wcschr(name, L'='))
 		continue;
-	    }
 	    unset_variable(name);
 	}
     }
 
-    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
+    return (yash_error_message_count == 0) ?
+	    Exit_SUCCESS : special_builtin_error(Exit_FAILURE);
 }
 
 /* Unsets the specified function.
@@ -2320,64 +2347,100 @@ const char unset_syntax[] = Ngt(
 );
 #endif
 
+/* Options for the "shift" built-in. */
+const struct xgetopt_T shift_options[] = {
+    { L'A', L"array", OPTARG_REQUIRED, false, NULL, },
+#if YASH_ENABLE_HELP
+    { L'-', L"help",  OPTARG_NONE,     false, NULL, },
+#endif
+    { L'\0', NULL, 0, false, NULL, },
+};
+
 /* The "shift" built-in */
 int shift_builtin(int argc, void **argv)
 {
+    const wchar_t *arrayname = NULL;
+
     const struct xgetopt_T *opt;
     xoptind = 0;
-    while ((opt = xgetopt(argv, help_option, 0)) != NULL) {
+    while ((opt = xgetopt(argv, shift_options, XGETOPT_DIGIT)) != NULL) {
 	switch (opt->shortopt) {
+	    case L'A':
+		arrayname = xoptarg;
+		break;
 #if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     if (!validate_operand_count(argc - xoptind, 0, 1))
-	return special_builtin_syntax_error(Exit_ERROR);
+	return special_builtin_error(Exit_ERROR);
 
-    size_t scount;
+    long count;
     if (xoptind < argc) {
-	long count;
 	if (!xwcstol(ARGV(xoptind), 10, &count)) {
 	    xerror(errno, Ngt("`%ls' is not a valid integer"), ARGV(xoptind));
-	    return special_builtin_syntax_error(Exit_ERROR);
-	} else if (count < 0) {
+	    return special_builtin_error(Exit_ERROR);
+	} else if (posixly_correct && count < 0) {
 	    xerror(0, Ngt("%ls: the operand value must not be negative"),
 		    ARGV(xoptind));
-	    return special_builtin_syntax_error(Exit_ERROR);
+	    return special_builtin_error(Exit_ERROR);
 	}
-#if LONG_MAX > SIZE_MAX
-	if (count > (long) SIZE_MAX)
-	    scount = SIZE_MAX;
-	else
-#endif
-	    scount = (size_t) count;
     } else {
-	scount = 1;
+	count = 1;
     }
 
-    variable_T *var = search_variable(L VAR_positional);
-    assert(var != NULL && (var->v_type & VF_MASK) == VF_ARRAY);
-    if (scount > var->v_valc) {
-	xerror(0,
-	    ngt("%zu: cannot shift so many "
-		"(there is only one positional parameter)",
-		"%zu: cannot shift so many "
-		"(there are only %zu positional parameters)",
-		var->v_valc),
-	    scount, var->v_valc);
+    variable_T *var;
+    if (arrayname == NULL) {
+	var = search_variable(L VAR_positional);
+	assert(var != NULL && (var->v_type & VF_MASK) == VF_ARRAY);
+    } else {
+	var = search_variable(arrayname);
+	if (wcschr(arrayname, L'=') != NULL ||
+		var == NULL || (var->v_type & VF_MASK) != VF_ARRAY) {
+	    xerror(0, Ngt("$%ls is not an array"), arrayname);
+	    return Exit_FAILURE;
+	}
+    }
+
+    unsigned long abscount =
+	(count >= 0) ? (unsigned long) count : -(unsigned long) count;
+
+    if (
+#if ULONG_MAX > SIZE_MAX
+	    abscount > (unsigned long) var->v_valc
+#else
+	    (size_t) abscount > var->v_valc
+#endif
+	    ) {
+	const char *message;
+	if (arrayname == NULL) {
+	    message = ngt("%ld: cannot shift so many "
+		    "(there is only one positional parameter)",
+		    "%ld: cannot shift so many "
+		    "(there are only %zu positional parameters)",
+		    var->v_valc);
+	} else {
+	    message = ngt("%ld: cannot shift so many "
+		    "(there is only one array element)",
+		    "%ld: cannot shift so many "
+		    "(there are only %zu array elements)",
+		    var->v_valc);
+	}
+	xerror(0, message, count, var->v_valc);
 	return Exit_FAILURE;
     }
 
+    size_t from = (count >= 0) ? 0 : (var->v_valc - (size_t) abscount);
     plist_T list;
     pl_initwith(&list, var->v_vals, var->v_valc);
-    for (size_t i = 0; i < scount; i++)
-	free(list.contents[i]);
-    pl_remove(&list, 0, scount);
+    for (size_t i = 0; i < (size_t) abscount; i++)
+	free(list.contents[from + i]);
+    pl_remove(&list, from, (size_t) abscount);
     var->v_valc = list.length;
     var->v_vals = pl_toary(&list);
 
@@ -2386,10 +2449,10 @@ int shift_builtin(int argc, void **argv)
 
 #if YASH_ENABLE_HELP
 const char shift_help[] = Ngt(
-"remove some positional parameters"
+"remove some positional parameters or array elements"
 );
 const char shift_syntax[] = Ngt(
-"\tshift [count]\n"
+"\tshift [-A array_name] [count]\n"
 );
 #endif
 
@@ -2546,7 +2609,7 @@ bool check_options(const wchar_t *options)
 		return false;
 	    /* falls thru! */
 	default:
-	    if (posixly_correct && !iswalpha(*options))
+	    if (posixly_correct && !iswalnum(*options))
 		return false;
 	    options++;
 	    if (*options == L':')
@@ -2616,7 +2679,13 @@ struct reading_option_T {
  */
 int read_builtin(int argc, void **argv)
 {
-    struct reading_option_T ro = { 0 };
+    struct reading_option_T ro = {
+	.array = false,
+	.lineedit = false,
+	.ps1 = false,
+	.raw = false,
+	.prompt = NULL,
+    };
 
     const struct xgetopt_T *opt;
     xoptind = 0;
@@ -2671,6 +2740,7 @@ int read_builtin(int argc, void **argv)
     }
 
     /* split fields */
+    const wchar_t *tail;
     plist_T list;
     pl_init(&list);
     {
@@ -2678,7 +2748,7 @@ int read_builtin(int argc, void **argv)
 	if (ifs == NULL)
 	    ifs = DEFAULT_IFS;
 
-	extract_fields(buf.contents, split.contents, false, ifs, &list);
+	tail = extract_fields(buf.contents, split.contents, false, ifs, &list);
 	assert(list.length % 2 == 0);
     }
 
@@ -2704,8 +2774,11 @@ int read_builtin(int argc, void **argv)
     if (ro.array) {
 	assign_array(name, &list, 2 * (count - 1));
     } else {
-	const wchar_t *start = list.contents[2 * (count - 1)];
-	const wchar_t *end = list.contents[list.length - 1];
+	size_t i = count - 1;
+	const wchar_t *start = list.contents[2 * i];
+	const wchar_t *end = list.contents[2 * i + 1];
+	if (2 * count < list.length)
+	    end = tail;
 	wchar_t *field = xwcsndup(start, end - start);
 	set_variable(name, field, SCOPE_GLOBAL, shopt_allexport);
     }
@@ -2886,7 +2959,7 @@ const char read_help[] = Ngt(
 "read a line from the standard input"
 );
 const char read_syntax[] = Ngt(
-"\tread [-Ar] variable...\n"
+"\tread [-Aer] [-P|-p] variable...\n"
 );
 #endif
 
@@ -3011,7 +3084,7 @@ variable_T *get_dirstack(void)
     variable_T *var = search_variable(L VAR_DIRSTACK);
     if (var != NULL) {
 	if ((var->v_type & VF_MASK) != VF_ARRAY) {
-	    xerror(0, Ngt("$DIRSTACK is not an array"));
+	    xerror(0, Ngt("$%ls is not an array"), L VAR_DIRSTACK);
 	    return NULL;
 	} else if (var->v_type & VF_READONLY) {
 	    xerror(0, Ngt("$%ls is read-only"), L VAR_DIRSTACK);
@@ -3033,7 +3106,7 @@ variable_T *get_dirstack(void)
 void push_dirstack(variable_T *var, wchar_t *value)
 {
     size_t index = var->v_valc++;
-    var->v_vals = xreallocn(var->v_vals, index + 2, sizeof *var->v_vals);
+    var->v_vals = xrealloce(var->v_vals, index, 2, sizeof *var->v_vals);
     var->v_vals[index] = value;
     var->v_vals[index + 1] = NULL;
 }
